/*
(c) Copyright 1994 HEWLETT-PACKARD COMPANY

To anyone who acknowledges that this file is provided 
"AS IS" without any express or implied warranty:
permission to use, copy, modify, and distribute this 
file for any purpose is hereby granted without fee, 
provided that the above copyright notice and this 
notice appears in all copies, and that the name of 
Hewlett-Packard Company not be used in advertising or 
publicity pertaining to distribution of the software 
without specific, written prior permission.  Hewlett-
Packard Company makes no representations about the 
suitability of this software for any purpose.
*/



#include "xscmlibint.h"

    /*  A 332 dither except dithers to 332 with the "Color Recovery"
        dither.  Should be used only to windows that have the smoothing cmap.

        NOTE: for more info on Color Recovery, access the Mosaic/WWW URL:
            http://hpfcdma.fc.hp.com/MHM/WhitePapers.hp/WhitePapers.html
        and under "White Papers" select "HP Color Recovery".
    */
int XscmRenderCR (RENDER_PARS)
    RENDER_PARS_FULL
{
int                red, green, blue;
int                x, y;
long               pixel;
unsigned char     *pSrc, *pDst;
unsigned char     *pDRed, *pDGreen, *pDBlue;
unsigned char     *pLineRed, *pLineGreen, *pLineBlue;

static short dither_red[2][16] = {
   {-16,  4, -1, 11,-14,  6, -3,  9,-15,  5, -2, 10,-13,  7, -4,  8},
   { 15, -5,  0,-12, 13, -7,  2,-10, 14, -6,  1,-11, 12, -8,  3, -9} };
static short dither_green[2][16] = {
   { 11,-15,  7, -3,  8,-14,  4, -2, 10,-16,  6, -4,  9,-13,  5, -1},
   {-12, 14, -8,  2, -9, 13, -5,  1,-11, 15, -7,  3,-10, 12, -6,  0} };
static short dither_blue[2][16] = {
   { -3,  9,-13,  7, -1, 11,-15,  5, -4,  8,-14,  6, -2, 10,-16,  4},
   {  2,-10, 12, -8,  0,-12, 14, -6,  3, -9, 13, -7,  1,-11, 15, -5} };
 
    /*  Lookup table for red/green/blue, for line%16 = 0 and 1: 
        indexed by r/g/b, for x%16 = 0..15.
    */
static unsigned char *pTable = (unsigned char *)NULL;
#define TABLE_RED_0    (&pTable[0 * (16*256)])
#define TABLE_GREEN_0  (&pTable[1 * (16*256)])
#define TABLE_BLUE_0   (&pTable[2 * (16*256)])
#define TABLE_RED_1    (&pTable[3 * (16*256)])
#define TABLE_GREEN_1  (&pTable[4 * (16*256)])
#define TABLE_BLUE_1   (&pTable[5 * (16*256)])

#define CLAMP(_comp) { \
    if (_comp > 0xff) _comp = 0xff; \
    else if (_comp < 0x00) _comp = 0x00; \
    }


        /*  Copy std tables (one time) to more efficient form. 
            malloc space for whole table and init it.  Table looks as follows:
              red/green/blue0 (line mod 2 == 0 tables)
              red/green/blue1 (line mod 2 == 1 tables)
            each table is essentially [16][256]: tables for x = 0..15 [16]
            and the right red/green/blue value for each 0..255 value  [256].
        */
    if (!pTable) {
        int i, comp;
        unsigned char *red0, *green0, *blue0;
        unsigned char *red1, *green1, *blue1;

        pTable = (unsigned char *)malloc(6 * 16 * 256);
        if (!pTable)
            return FALSE;
        red0   = TABLE_RED_0;
        green0 = TABLE_GREEN_0;
        blue0  = TABLE_BLUE_0;
        red1   = TABLE_RED_1;
        green1 = TABLE_GREEN_1;
        blue1  = TABLE_BLUE_1;

        for (x = 0; x < 16; x++) {
            for (i = 0; i < 256; i++) {
                comp = i;
                if (comp >= 48)
                     comp = comp - 16;
                else comp = comp / 2 + 8;

                red = comp + dither_red[0][x];
                CLAMP (red);
                red0[(x * 256) + i] = red;
                red = comp + dither_red[1][x];
                CLAMP (red);
                red1[(x * 256) + i] = red;

                green = comp + dither_green[0][x];
                CLAMP (green);
                green0[(x * 256) + i] = green;
                green = comp + dither_green[1][x];
                CLAMP (green);
                green1[(x * 256) + i] = green;

                comp = i;
                if (comp >= 112)
                     comp = comp - 32;
                else comp = comp / 2 + 24;

                blue += dither_blue[0][x];
                CLAMP (blue);
                blue0[(x * 256) + i] = green;
                blue += dither_blue[1][x];
                CLAMP (blue);
                blue1[(x * 256) + i] = green;
	        }
	    }
        }


    for (y = 0; y < height; y++) {
        if (y & 1) {
            pLineRed   = TABLE_RED_1;
            pLineGreen = TABLE_GREEN_1;
            pLineBlue  = TABLE_BLUE_1;
	    }
        else  {
            pLineRed   = TABLE_RED_0;
            pLineGreen = TABLE_GREEN_0;
            pLineBlue  = TABLE_BLUE_0;
	    }
        pSrc = src;
        src += srcRowBytes;
        pDst = dst;
        dst += dstRowBytes;

        for (x = 0; x < width; x++) {
            if ((x & 0xf) == 0) {
                pDRed   = pLineRed;
                pDGreen = pLineGreen;
                pDBlue  = pLineBlue;
	        }
            else  {
                pDRed   += 256;
                pDGreen += 256;
                pDBlue  += 256;
	        }

            if (pRed) {
                pixel = *pSrc++;
                red   = pRed[pixel] >> 8;
                green = pGreen[pixel] >> 8;
                blue  = pBlue[pixel] >> 8;
	        }
            else {
                red   = *pSrc++;
                green = *pSrc++;
                blue  = *pSrc++;
	        }
            red   = pDRed[red];
            green = pDGreen[green];
            blue  = pDBlue[blue];
            pixel = ((red & 0xE0) | ((green & 0xE0) >> 3) 
                                  | ((blue & 0xC0) >> 6));
            *pDst++ = pixel;
            }
        }
    return TRUE;
}


DISPLAY_FUNCTION (XscmDisplayCR, XscmRenderCR)

