/*
(c) Copyright 1994 HEWLETT-PACKARD COMPANY

To anyone who acknowledges that this file is provided 
"AS IS" without any express or implied warranty:
permission to use, copy, modify, and distribute this 
file for any purpose is hereby granted without fee, 
provided that the above copyright notice and this 
notice appears in all copies, and that the name of 
Hewlett-Packard Company not be used in advertising or 
publicity pertaining to distribution of the software 
without specific, written prior permission.  Hewlett-
Packard Company makes no representations about the 
suitability of this software for any purpose.
*/



#include "xscmlibint.h"


    /*  Dither pixels from "src" to "dst" (232 bit RGB byte/pixel).
        If "pRed" is null, "src" points to an array of RGB values:
        each a byte, in that order; "srcRowBytes" is the # bytes between
        rows (aka stride).  "dstRowBytes" is bytes/row of "dst".
        If "pRed" is non-null, then "src" is taken as an array of bytes,
        each an index into pRed/pGreen/pBlue, which are *16 bit* color
        values; in other words pseudo-color with X-like color values.
    */
int XscmRenderDefault (RENDER_PARS)
    RENDER_PARS_FULL
{
int                basePixel = pInfo->basePixel;
int                red, green, blue;
int                redAErr, greenAErr, blueAErr;
int                redALErr, greenALErr, blueALErr;
int                redARErr, greenARErr, blueARErr;
long               pixel;
int               *pErr;
int                thisRed, thisGreen, thisBlue;
unsigned char     *pSrc, *pDst;
long               nPixelsM1;
#define            LOCAL_WIDTH 1024
int                errs[LOCAL_WIDTH], *pErrs;
int                index, i, redLevel, greenLevel, blueLevel;
static long        ColorTable[128];
static int         tableInited = 0;

    if (!tableInited) {
        tableInited = 1;
        for (index = 0, red = 0; red < 4; red++) {
            redLevel = 255 * red / 3;
            for (green = 0; green < 8; green++) {
                greenLevel = 255 * green / 7;
                for (blue = 0; blue < 4; blue++) {
                    blueLevel = 255 * blue / 3;
                    i = (blueLevel << 16) | (greenLevel << 8) | redLevel;
                    ColorTable[index] = i;
                    index++;
                    }
                }
            }
        }

    pixel = (width + 6) * 3;  /* # of elements in error array */
    if (pixel <= LOCAL_WIDTH)
        pErrs = errs;
    else pErrs = (int *)malloc(pixel * sizeof(int));
    for (pErr = pErrs, red = 0; red < pixel; red++)
        *pErr++ = 0;

    while (height-- > 0) {
        pSrc = src;
        src += srcRowBytes;
        pDst = dst;
        dst += dstRowBytes;
        nPixelsM1 = width - 1;

        pErr = pErrs;
        red = redAErr = 0;
        redARErr = *pErr;
        green = greenAErr = 0;
        greenARErr = *(pErr + 1);
        blue = blueAErr = 0;
        blueARErr = *(pErr + 2);

        do {
            redALErr = redAErr;
            redAErr = redARErr;
            redARErr = *(pErr + 3);

            /* red,green,blue are errors from previous pixel to left.  Get
               color of this pixel, fetched from color tables (16->8 bit).
            */
            if (pRed) {
                pixel = *pSrc++;
                thisRed   = pRed[pixel] >> 8;
                thisGreen = pGreen[pixel] >> 8;
                thisBlue  = pBlue[pixel] >> 8;
	        }
            else {
                thisRed   = *pSrc++;
                thisGreen = *pSrc++;
                thisBlue  = *pSrc++;
	        }

            pixel = red;
            red += redAErr;
            pixel += red;
            red += redARErr;
            pixel += red;
            red = thisRed + ((pixel + pixel + red + redALErr) >> 4);
            if (red >> 8) goto edClipR;
edClipRetR:
            greenALErr = greenAErr;
            greenAErr = greenARErr;
            greenARErr = *(pErr + 4);
            pixel = green;
            green += greenAErr;
            pixel += green;
            green += greenARErr;
            pixel += green;
            green = thisGreen + ((pixel + pixel + green + greenALErr) >> 4);
            if (green >> 8) goto edClipG;
edClipRetG:
            blueALErr = blueAErr;
            blueAErr = blueARErr;
            blueARErr = *(pErr + 5);
            pixel = blue;
            blue += blueAErr;
            pixel += blue;
            blue += blueARErr;
            pixel += blue;
            blue = thisBlue + ((pixel + pixel + blue + blueALErr) >> 4);
            if (blue >> 8) goto edClipB;
edClipRetB:

            /* red,green,blue now contain desired color (0..255). Pack to
               (RGB bits) 232; set each component to error = desired color
               (component) - actual (from byte in SCMColorTable).
            */
            pixel = red >> 6;  /* get top bits */
            redALErr = (green >> 5);
            pixel = (pixel << 3) + redALErr;
            redALErr = (blue >> 6);
            pixel = (pixel << 2) + redALErr;
            *pDst++ = (unsigned char)(pixel + basePixel);

            pixel = ColorTable[pixel];
            red -= ((unsigned char)pixel);
            *pErr++ = red;
            pixel >>= 8;
            green -= ((unsigned char)pixel);
            *pErr++ = green;
            blue -= (unsigned char)(pixel >> 8);
            *pErr++ = blue;
            } while (--nPixelsM1 >= 0);
        }     /* END while height; one scan line */

    if (pErrs != errs)     /* not local var; free it */
        free(pErrs);

    return TRUE;

edClipR:   if (red < 0) red = 0; else red = 255;
            goto edClipRetR;
edClipG:   if (green < 0) green = 0; else green = 255;
            goto edClipRetG;
edClipB:   if (blue < 0) blue = 0; else blue = 255;
            goto edClipRetB;
}


DISPLAY_FUNCTION (XscmDisplayDefault, XscmRenderDefault)

