/*
(c) Copyright 1994 HEWLETT-PACKARD COMPANY

To anyone who acknowledges that this file is provided 
"AS IS" without any express or implied warranty:
permission to use, copy, modify, and distribute this 
file for any purpose is hereby granted without fee, 
provided that the above copyright notice and this 
notice appears in all copies, and that the name of 
Hewlett-Packard Company not be used in advertising or 
publicity pertaining to distribution of the software 
without specific, written prior permission.  Hewlett-
Packard Company makes no representations about the 
suitability of this software for any purpose.
*/



#include "xscmlibint.h"

    /*  Given a visual id return its info */
static void XscmGetVisualInfo (display, visualid, pInfo)
    Display        *display;
    int             visualid;
    XVisualInfo    *pInfo;
{
XVisualInfo         vinfo, *vptr;
int                 nItems;

    vinfo.visualid = visualid;
    vptr = XGetVisualInfo (display, VisualIDMask, &vinfo, &nItems);
    *pInfo = *vptr;
    XFree ((char *)vptr);
}


int XscmFind24BitSCM (display, screen, pInfo)
    Display        *display;
    int             screen;
    XscmInfoRec    *pInfo;
{
Window          root;
XStandardColormap *pStd;
int             nStd, i;
XVisualInfo      vinfo;

    root = RootWindow (display, screen);
    if (XGetRGBColormaps (display, root, &pStd, &nStd, XA_RGB_BEST_MAP)) {
        XscmGetVisualInfo (display, pStd->visualid, &vinfo);

        if ((vinfo.depth == 24)
         && (pStd->red_max == 255)
         && (pStd->green_max == 255)
         && (pStd->blue_max == 255)) {
            pInfo->redShift = pInfo->greenShift = pInfo->blueShift = 0;
            i = pStd->red_mult;
            while ((i & 1) == 0) {i >>= 1; pInfo->redShift +=1;}
            if (i != 1) return FALSE;
            i = pStd->green_mult;
            while ((i & 1) == 0) {i >>= 1; pInfo->greenShift +=1;}
            if (i != 1) return FALSE;
            i = pStd->blue_mult;
            while ((i & 1) == 0) {i >>= 1; pInfo->blueShift +=1;}
            if (i != 1) return FALSE;

            pInfo->renderType = XSCM_RENDER_24_BIT;
            pInfo->display = display;
            pInfo->screen = screen;
            pInfo->visual = vinfo.visual;
            pInfo->colormap = pStd->colormap;
            pInfo->depth = vinfo.depth;
            pInfo->bytesPerPixel = 4;
            pInfo->basePixel = pStd->base_pixel;
            return TRUE;
	    }
        }
    return FALSE;
}


int XscmFind8BitSCM (display, screen, pInfo)
    Display        *display;
    int             screen;
    XscmInfoRec    *pInfo;
{
Window          root;
XStandardColormap *pStd;
int             nStd, i;
XVisualInfo      vinfo;

    root = RootWindow (display, screen);
    if (XGetRGBColormaps (display, root, &pStd, &nStd, XA_RGB_BEST_MAP)) {
        XscmGetVisualInfo (display, pStd->visualid, &vinfo);

        if ((vinfo.depth == 8)
         && (pStd->red_max == 7)   && (pStd->red_mult == 32)
         && (pStd->green_max == 7) && (pStd->green_mult == 4) 
         && (pStd->blue_max == 3)  && (pStd->blue_mult == 1)) {
            pInfo->renderType = XSCM_RENDER_884;
            pInfo->display = display;
            pInfo->screen = screen;
            pInfo->visual = vinfo.visual;
            pInfo->colormap = pStd->colormap;
            pInfo->depth = vinfo.depth;
            pInfo->bytesPerPixel = 1;
            pInfo->basePixel = pStd->base_pixel;
            return TRUE;
	    }
        }
    return FALSE;
}


int XscmFindDefaultSCM (display, screen, pInfo)
    Display        *display;
    int             screen;
    XscmInfoRec    *pInfo;
{
Window          root;
XStandardColormap *pStd;
int             nStd, i;
XVisualInfo      vinfo;

    root = RootWindow (display, screen);
    if (XGetRGBColormaps (display, root, &pStd, &nStd, XA_RGB_DEFAULT_MAP)) {
        XscmGetVisualInfo (display, pStd->visualid, &vinfo);

        if ((vinfo.depth == 8)
         && (pStd->red_max == 3)   && (pStd->red_mult == 32)
         && (pStd->green_max == 7) && (pStd->green_mult == 4) 
         && (pStd->blue_max == 3)  && (pStd->blue_mult == 1)) {
            pInfo->renderType = XSCM_RENDER_484;
            pInfo->display = display;
            pInfo->screen = screen;
            pInfo->visual = vinfo.visual;
            pInfo->colormap = pStd->colormap;
            pInfo->depth = vinfo.depth;
            pInfo->bytesPerPixel = 1;
            pInfo->basePixel = pStd->base_pixel;
            return TRUE;
	    }
        }
    return FALSE;
}


int XscmFindGraySCM (display, screen, pInfo)
    Display        *display;
    int             screen;
    XscmInfoRec    *pInfo;
{
Window          root;
XStandardColormap *pStd;
int             nStd, i;
XVisualInfo      vinfo;

    root = RootWindow (display, screen);
    if (XGetRGBColormaps (display, root, &pStd, &nStd, XA_RGB_GRAY_MAP)) {
        XscmGetVisualInfo (display, pStd->visualid, &vinfo);
        if ((vinfo.visual == DefaultVisual (display, screen))
         && (vinfo.depth == 8)
         && (pStd->red_max == 31)) {
            pInfo->renderType = XSCM_RENDER_GRAY;
            pInfo->display = display;
            pInfo->screen = screen;
            pInfo->visual = DefaultVisual (display, screen);
            pInfo->colormap = pStd->colormap;
            pInfo->depth = 8;
            pInfo->bytesPerPixel = 1;
            pInfo->basePixel = pStd->base_pixel;
            pInfo->redShift = pInfo->greenShift = pInfo->blueShift = 0;
            return TRUE;
	    }
        }
    return FALSE;
}


int XscmFindCRSCM (display, screen, useDefaultVisual, pInfo)
    Display        *display;
    int             screen;
    int             useDefaultVisual;
    XscmInfoRec    *pInfo;
{
Window              root;
XStandardColormap  *pStd;
int                 nStd, i, found, desiredVisualID;
Visual             *visual;
Atom                atom;
XVisualInfo         vinfo;

        /*  Find Color Recovery map: if useDefaultVisual get one that has
            same visual as default (will be PseudoColor) else get the 
            TrueColor visual.
        */
    found = FALSE;
    root = RootWindow (display, screen);
    if (useDefaultVisual)
        visual = DefaultVisual(display, screen);
    else {
        if (!XMatchVisualInfo (display, screen, 8, TrueColor, &vinfo))
            return FALSE;
        visual = vinfo.visual;
        }
    desiredVisualID = visual->visualid;

        /*  Find Color Recovery smooth map, using pre-9.05 property. */
    atom = XInternAtom (display, "_HP_RGB_SMOOTH_PSEUDO_MAP", 1);
    if ((atom != None) 
     && XGetRGBColormaps (display, root, &pStd, &nStd, atom)
     && (pStd->visualid == desiredVisualID))
        found = TRUE;

    if (!found) {
        atom = XInternAtom (display, "_HP_RGB_SMOOTH_TRUE_MAP", 1);
        if ((atom != None) 
         && XGetRGBColormaps (display, root, &pStd, &nStd, atom)
         && (pStd->visualid == desiredVisualID))
            found = TRUE;
        }

        /*  If not, find list; pick the one with desiredVisualID. */
    if (!found) {
        nStd = 0;
        atom = XInternAtom (display, "_HP_RGB_SMOOTH_MAP_LIST", 1);
        if ((atom != None) 
         && XGetRGBColormaps(display, root, &pStd, &nStd, atom)) {
            while (1) {
                if (pStd->visualid == desiredVisualID) {
                    found = 1;
                    break;
		    }
                if (--nStd <= 0)
                    return 0;
                pStd++;
    	        }
            }
        }

    if (found) {
        XscmGetVisualInfo (display, pStd->visualid, &vinfo);
        if ((vinfo.depth == 8)
         && (pStd->red_max == 7)   && (pStd->red_mult == 32)
         && (pStd->green_max == 7) && (pStd->green_mult == 4) 
         && (pStd->blue_max == 3)  && (pStd->blue_mult == 1)) {
            pInfo->renderType = XSCM_RENDER_884_CR;
            pInfo->display = display;
            pInfo->screen = screen;
            pInfo->visual = visual;
            pInfo->colormap = pStd->colormap;
            pInfo->depth = 8;
            pInfo->bytesPerPixel = 1;
            pInfo->basePixel = 0;
            pInfo->redShift = pInfo->greenShift = pInfo->blueShift = 0;
            return TRUE;
            }
        }
    return FALSE;
}


int XscmFindColorSCM (display, screen, type, pInfo)
    Display        *display;
    int             screen;
    int             type;
    XscmInfoRec    *pInfo;
{
Visual             *defaultVisual;
Colormap            defaultColormap;

        /*  Set defaultVisual/Colormap to null if not required, else init. */
    defaultVisual = (Visual *)NULL;
    defaultColormap = (Colormap)0;
    switch (type) {
      case XSCM_ANY:
        break;
      case XSCM_DEFAULT_COLORMAP:
        defaultColormap = DefaultColormap (display, screen);
      case XSCM_DEFAULT_VISUAL:
        defaultVisual = DefaultVisual (display, screen);
        break;
        }

        /*  Find "best": 24 bit, default visual/colormap if requested. */
    if (XscmFind24BitSCM (display, screen, pInfo)) {
        if ((!defaultVisual || (pInfo->visual == defaultVisual))
         && (!defaultColormap || (pInfo->colormap == defaultColormap)))
        return TRUE;
        }

        /*  Find Color Recovery colormap unless must use default colormap. 
            Use default visual if caller requires it.
        */
    if (!defaultColormap) {
        if (XscmFindCRSCM (display, screen, (defaultVisual != (Visual *)NULL),
                           pInfo))
            return TRUE;
        }

        /*  Find 8 bit TrueColor visual */
    if (XscmFind8BitSCM (display, screen, pInfo)) {
        if ((!defaultVisual || (pInfo->visual == defaultVisual))
         && (!defaultColormap || (pInfo->colormap == defaultColormap)))
        return TRUE;
        }

        /*  Find default map; by definition uses default visual/colormap. */
    if (XscmFindDefaultSCM (display, screen, pInfo)) {
        if ((!defaultVisual || (pInfo->visual == defaultVisual))
         && (!defaultColormap || (pInfo->colormap == defaultColormap)))
        return TRUE;
        }

    return FALSE;
}

