/*
(c) Copyright 1994 HEWLETT-PACKARD COMPANY

To anyone who acknowledges that this file is provided 
"AS IS" without any express or implied warranty:
permission to use, copy, modify, and distribute this 
file for any purpose is hereby granted without fee, 
provided that the above copyright notice and this 
notice appears in all copies, and that the name of 
Hewlett-Packard Company not be used in advertising or 
publicity pertaining to distribution of the software 
without specific, written prior permission.  Hewlett-
Packard Company makes no representations about the 
suitability of this software for any purpose.
*/



#include <stdio.h>
#include <string.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Xatom.h>

#define FALSE 0
#define TRUE  1
#define NULL  0


int XscmFindBest (display, screen, pStd)
    Display    *display;
    int         screen;

    /*  Create the given std cmap: check if already defined; if so, delete
        it first if -replace specified.  Return true if one was created.
    */
static int CreateOne (display, screen, visualid, depth, colormap,
                      redMax, greenMax, blueMax, property, doReplace)
    Display    *display;
    int         screen;
    int         visualid;
    int         depth;
    Colormap    colormap;
    int         redMax;
    int         greenMax;
    int         blueMax;
    Atom        property;
    int         doReplace;
{
XStandardColormap std, *pStd;
int             i;
Window          root = RootWindow (display, screen);

    if (XGetRGBColormaps (display, root, &std, &i, property)) {
        if (doReplace)
            XmuDeleteStandardColormap (display, screen, property);
        else {
            fprintf (stderr, "? std cmap exists; need -replace.\n");
            return FALSE;
            }
        }

    pStd = XmuStandardColormap (display, screen, visualid, depth, property,
                              colormap, redMax, greenMax, blueMax);
    if (!pStd) {
        fprintf (stderr, "? error creating std colormap.\n");
        return FALSE;
        }
    XSetRGBColormaps (display, root, pStd, 1, property);
    return TRUE;
}



    /*  Create XA_RGB_BEST_MAP on given screen; return true iff created. */
static int CreateBest (display, screen, nVisuals, pVisuals, doReplace)
    Display    *display;
    int         screen;
    int         nVisuals;
    XVisualInfo *pVisuals;
    int         doReplace;
{
Visual         *visual;
int             class;
XVisualInfo    *pVisual, *pBest;
int             redMax, greenMax, blueMax;

        /*  Find "best" visual to use, based on following rules:
               - don't use DefaultVisual
               - highest depth visual used
               - True/Direct/Static/PseudoColor used in that order; no gray!
        */
    pBest = (XVisualInfo *)NULL;
    for (pVisual = pVisuals; nVisuals-- > 0; pVisual++) {
        visual = pVisual->visual;
        if (visual == DefaultVisual (display, screen))
            continue;

        class = pVisual->class;
        if ((class != TrueColor) && (class != DirectColor) 
         && (class != StaticColor) && (class != PseudoColor))
            continue;

            /*  Visual is eligible; if first one note and continue. */
        if (!pBest) {
            pBest = pVisual;
            continue;
	    }

            /*  Not first one; set this one if better than previous best. */
        if (pVisual->depth > pBest->depth) {
            pBest = pVisual;
            continue;
	    }
        if (pVisual->depth < pBest->depth)
            continue;

            /*  Depth same; choose based on descending desireable classes. */
        switch (class) {
          case TrueColor:
            pBest = pVisual;
            break;
          case DirectColor:
            if (pBest->class != TrueColor)
                pBest = pVisual;
            break;
          case StaticColor:
            if (pBest->class == PseudoColor)
                pBest = pVisual;
            break;
          case PseudoColor:
            break;
            }
        }                                /* END for; find visual */

        /*  Setup "std" with colormap values to set. */
    if (pBest) {
        switch (pBest->class) {

              /*  For True/DirectColor allocate all available RGBs: max of
                  each component = the mask, aligned to low order.
              */
          case TrueColor:
          case DirectColor:
            redMax = pBest->red_mask;
            while ((redMax & 1) == 0) redMax >>= 1;
            greenMax = pBest->green_mask;
            while ((greenMax & 1) == 0) greenMax >>= 1;
            blueMax = pBest->blue_mask;
            while ((blueMax & 1) == 0) blueMax >>= 1;
            break;

          case StaticColor:
          case PseudoColor:
            pBest = (XVisualInfo *)NULL; /* not supported; no best one */
            break;
            }
        }

        /*  If best exists create std map. */
    if (pBest)
        return CreateOne (display, screen, pBest->visualid, pBest->depth, 
                         (Colormap)None, redMax, greenMax, blueMax, 
                         XA_RGB_BEST_MAP, doReplace);
    else return FALSE;
}



    /*  Create XA_RGB_DEFAULT_MAP on given screen; return true iff created. */
static int CreateDefault (display, screen, doReplace)
    Display    *display;
    int         screen;
    int         doReplace;
{
#define RMAX    3           /* RGB max for a 484 colormap */
#define GMAX    7
#define BMAX    3

    return CreateOne (display, screen, 
       DefaultVisual (display, screen)->visualid, 
       DefaultDepth (display, screen), DefaultColormap (display, screen), 
       RMAX, GMAX, BMAX, XA_RGB_DEFAULT_MAP, doReplace);
}


    /*  Create XA_RGB_GRAY_MAP on given screen; return true iff created. */
static int CreateGray (display, screen, doReplace)
    Display    *display;
    int         screen;
    int         doReplace;
{
#define GRAYMAX 31         /* # of grays to allocate - 1 */

        /*  Very hacky, but it works.  Xmu stuff insists that green/blue max
            can't be zero or error; but the total # colors allocated is the
            sum of all, but want 32 red shades (how gray map is used).
            So ask for -1/1 green/blue to balance each other out.
        */
    return CreateOne (display, screen, 
       DefaultVisual (display, screen)->visualid, 
       DefaultDepth (display, screen), DefaultColormap (display, screen), 
       GRAYMAX, -1, 1, XA_RGB_GRAY_MAP, doReplace);
}


    /*  Print info from given std cmap info. */
static void PrintOne (pStd)
    XStandardColormap *pStd;
{
    printf ("    Colormap: 0x%x\n", pStd->colormap);
    printf ("    red max %d mult %d\n", 
                 pStd->red_max, pStd->red_mult);
    printf ("    green max %d mult %d\n", 
                 pStd->green_max, pStd->green_mult);
    printf ("    blue max %d mult %d\n", 
                 pStd->blue_max, pStd->blue_mult);
    printf ("    base pixel %d (0x%x) \n", 
                 pStd->base_pixel, pStd->base_pixel);
    printf ("    Visual id: 0x%x\n", pStd->visualid);
    printf ("    Kill id: 0x%x\n", pStd->killid);
}


    /*  List best and default std cmaps found on all screens. */
static void List (display)
    Display    *display;
{
XVisualInfo    *pVisuals, vinfo;
int             nVisuals;
int             screen;
Window          root;
XStandardColormap *pStd;
int             nStd, i;

    for (screen = 0; screen < ScreenCount(display); screen++) {
        printf ("Std cmaps for screen %d:\n", screen);
        root = RootWindow (display, screen);
        vinfo.screen = screen;
        pVisuals = XGetVisualInfo (display, VisualScreenMask, &vinfo, 
                                   &nVisuals);
        if (!pVisuals || (nVisuals <= 0))
            continue;               /* ? no visuals - try next screen */

        if (XGetRGBColormaps (display, root, &pStd, &nStd, XA_RGB_BEST_MAP)) {
            for (i = 0; i < nStd; i++) {
                printf ("  best map (XA_RGB_BEST_MAP) (%d):\n", i);
                PrintOne (pStd);
                }
            }

        if (XGetRGBColormaps (display, root,&pStd,&nStd, XA_RGB_DEFAULT_MAP)) {
            for (i = 0; i < nStd; i++) {
                printf ("  best map (XA_RGB_DEFAULT_MAP) (%d):\n", i);
                PrintOne (pStd);
                }
            }

        if (XGetRGBColormaps (display, root,&pStd,&nStd, XA_RGB_GRAY_MAP)) {
            for (i = 0; i < nStd; i++) {
                printf ("  best map (XA_RGB_GRAY_MAP) (%d):\n", i);
                PrintOne (pStd);
                }
            }
        XFree (pVisuals);
        }
}



static void PrintUsage ()
{
    printf ("Usage: xscm [-help] [-display <dsp>]\n");
    printf ("    [-default] [-best] [-gray] [-all]\n");
    printf ("    [-replace] [-delete] [-list]\n");
    printf ("Normal usage: xscm -all -replace\n");
    printf ("\n");
}

static void PrintHelp ()
{
    PrintUsage();
    printf ("xscm creates, lists or deletes standard colormaps as used by\n");
    printf ("XGetRGBColormaps() on all screens.  Options:\n");
    printf ("\n");
    printf ("    -help          Print this message.\n");
    printf ("    -display <dsp> Use the X server named <dsp>\n");
    printf ("    -default       Create XA_RGB_DEFAULT_MAP in the default\n");
    printf ("                   colormap.  Uses up to 128 colormap slots.\n");
    printf ("    -gray          Create XA_RGB_GRAY_MAP in the default\n");
    printf ("                   colormap.  Uses up to 32 colormap slots.\n");
    printf ("    -best          Create XA_RGB_BEST_MAP, only if a TrueColor\n");
    printf ("                   or DirectColor visual exists.\n");
    printf ("    -all           Equivalent to '-default -gray -best'.\n");
    printf ("    -replace       Replace the standard colormap if it already\n");
    printf ("                   exists; default is to not replace it.\n");
    printf ("    -delete        Delete rather than create the default (if \n");
    printf ("                   -default) or best (if -best) colormaps.\n");
    printf ("    -list          List info on the standard colormaps found\n");
    printf ("                   after all other operations are done.\n");
}


main (argc, argv)
    int		argc;
    char	**argv;
{
int doReplace = FALSE;
int doDefault = FALSE;
int doGray = FALSE;
int doBest = FALSE;
int doDelete = FALSE;
int doList = FALSE;
char *displayName = "";
Display        *display = (Display *)NULL;

int             i;
char           *name;
#define STREQ(str, cmpstr) (strcmp ((str), (cmpstr)) == 0)

    argc--;
    argv++;

    if (argc <= 0) {
        PrintUsage ();
        exit (0);
        }

    while (argc-- > 0) {
        name = *argv++;
        if (STREQ (name, "-display")) {
            if (argc-- <= 0) {
                fprintf (stderr, "-display requires display name argument.\n");
                exit (1);
	        }
            displayName = *argv++;
	    }
        else if (STREQ (name, "-replace"))
            doReplace = TRUE;
        else if (STREQ (name, "-default"))
            doDefault = TRUE;
        else if (STREQ (name, "-gray"))
            doGray = TRUE;
        else if (STREQ (name, "-grey"))  /* for those who spell "colour" */
            doGray = TRUE;
        else if (STREQ (name, "-best"))
            doBest = TRUE;
        else if (STREQ (name, "-all")) {
            doDefault = TRUE;
            doGray = TRUE;
            doBest = TRUE;
	    }
        else if (STREQ (name, "-delete"))
            doDelete = TRUE;
        else if (STREQ (name, "-list"))
            doList = TRUE;
        else if (STREQ (name, "-help")) {
            PrintHelp ();
            exit (0);
            }
        else {
            fprintf (stderr, "? '%s' is not a valid option\n", name);
            exit (1);
	    }
        }

    if (doBest || doDefault || doGray || doDelete || doList) {
        display = XOpenDisplay (displayName);
        if (!display) {
            fprintf (stderr, "? Can't open display '%s'.\n", displayName);
            exit (1);
            }
        }

    if (doBest || doDefault || doGray) {
        XVisualInfo    *pVisuals, vinfo;
        int             nVisuals;
        int             screen;

        XSetCloseDownMode (display, RetainPermanent);
        for (screen = 0; screen < ScreenCount(display); screen++) {
            vinfo.screen = screen;
            pVisuals = XGetVisualInfo (display, VisualScreenMask, &vinfo, 
                                       &nVisuals);
            if (!pVisuals || (nVisuals <= 0))
                continue;           /* ? no visuals - try next screen */

            if (doBest) {
                if (doDelete)
	            XmuDeleteStandardColormap (display, screen, 
                                               XA_RGB_BEST_MAP);
	        else
                    CreateBest (display, screen, nVisuals, pVisuals, doReplace);
                }
            if (doDefault) {
                if (doDelete)
	            XmuDeleteStandardColormap (display, screen, 
                                               XA_RGB_DEFAULT_MAP);
	        else
                    CreateDefault (display, screen, doReplace);
                }
            if (doGray) {
                if (doDelete)
	            XmuDeleteStandardColormap (display, screen, 
                                               XA_RGB_GRAY_MAP);
	        else
                    CreateGray (display, screen, doReplace);
                }
            XFree (pVisuals);
            XSync (display, 0);
	    }
        }
    else if (doDelete) fprintf (stderr, 
        "? -delete requires a colormap (e.g. -default) to be specified.\n");

    if (doList)
        List (display);

    if (display)
        XCloseDisplay (display);
    exit (0);
}

