/*
(c) Copyright 1994 HEWLETT-PACKARD COMPANY

To anyone who acknowledges that this file is provided 
"AS IS" without any express or implied warranty:
permission to use, copy, modify, and distribute this 
file for any purpose is hereby granted without fee, 
provided that the above copyright notice and this 
notice appears in all copies, and that the name of 
Hewlett-Packard Company not be used in advertising or 
publicity pertaining to distribution of the software 
without specific, written prior permission.  Hewlett-
Packard Company makes no representations about the 
suitability of this software for any purpose.
*/



#include <stdio.h>
#include <string.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Xatom.h>

#include "xscmlib.h"

/*  ========================= SAMPLE CODE ================================= */
/*  This code not used in this program; it is here as sample code to be used
    in using Xscm, or in creating colormaps compatible with it.
*/

    /*  Given an X color definition map to nearest color in the Xscm "default"
        (484) map.  color->pixel is set to the *relative* value for the pixel,
        i.e. "base_pixel" in the XStandardColormap struct must be added to it
        to get the pixel to put in the frame buffer.   color.red/green/blue on
        entry contain the desired color; on return they contain the actual
        color of that pixel, i.e. if you then called XAllocColor() with "color"
        you would hit one of the 484 standard colors.
    */
void GetXscmDefaultPixel (color)
    XColor                 *color;
{
#define MAXRED              3
#define MAXGREEN            7
#define MAXBLUE             3
#define MAXDIV(_max)        ((double)65535 / (double)(_max))
int                         red, green, blue;

        /*  First find the nearest value for each component. */
    red   = (int)(((double)color->red / MAXDIV(MAXRED)) + 0.5);
    if (red > MAXRED) red = MAXRED;
    green = (int)(((double)color->green / MAXDIV(MAXGREEN)) + 0.5);
    if (green > MAXGREEN) green = MAXGREEN;
    blue  = (int)(((double)color->blue / MAXDIV(MAXBLUE)) + 0.5);
    if (blue > MAXBLUE) blue = MAXBLUE;

        /*  Pack into a pixel and return it. */
    color->pixel = (red * 32) + (green * 4) + blue;

        /*  Return actual colors that result. */
    color->red = red * 65535 / MAXRED;
    color->green = green * 65535 / MAXGREEN;
    color->blue = blue * 65535 / MAXBLUE;
}



    /*  The following allocates exactly the same colors as are used by 
        the Xscm 484 default map, e.g. for programs who do not directly use the
        484 map but want to use the same colors.
    */
int AllocateXscm484Colors (display, colormap)
    Display                *display;
    Colormap                colormap;
{
#define NRED                4
#define NGREEN              8
#define NBLUE               4
XColor                      color[128];
long                        i, red, green, blue;

    for (i = 0, red = 0; red < NRED; red++)
      for (green = 0; green < NGREEN; green++)
        for (blue = 0; blue < NBLUE; blue++, i++) {
            color[i].flags = DoRed | DoGreen | DoBlue;
            color[i].red   = red * 65535 / (NRED - 1); 
            color[i].green = green * 65535 / (NGREEN - 1); 
            color[i].blue  = blue * 65535 / (NBLUE - 1); 
            if (!XAllocColor (display, colormap, &color[i])) {
                while (--i > 0)
                    XFreeColors (display,colormap,&color[i].pixel,1,0);
                return 0;
                }
	    }
    return 1;
}


    /*  Given a shade of gray as defined for X (0..65535) map to nearest gray
        in the Xscm "gray" (32 shade) map. The *relative* value for the pixel
        is returned, i.e. "base_pixel" in the XStandardColormap struct must be 
        added to it to get the pixel to put in the frame buffer.   *pGray on
        entry contains the desired gray; on return it contains the actual
        gray of that pixel, i.e. if you then called XAllocColor() with "*pGray"
        in red/green/blue you would hit one of the 32 standard grays.
    */
unsigned long GetXscmGrayPixel (pGray)
    unsigned short          *pGray;

{
#define MAXGRAY             31
#define GRAYDIV             ((double)65535 / (double)(MAXGRAY))
int                         gray;

    gray = (int)(((double)*pGray / GRAYDIV) + 0.5);
    if (gray > MAXGRAY) gray = MAXGRAY;
    *pGray = gray  * 65535 / MAXGRAY;
    return gray;
}


    /*  The following allocates exactly the same colors (grays) as are used by 
        the Xscm 32 shade gray map.
    */
int AllocateXscm32Grays (display, colormap)
    Display                *display;
    Colormap                colormap;
{
#define NGRAYS              32
XColor                      color[NGRAYS];
long                        i;

    for (i = 0; i < NGRAYS; i++) {
        color[i].flags = DoRed | DoGreen | DoBlue;
        color[i].red = color[i].green = color[i].blue =
            i * 65535 / (NGRAYS - 1); 
        if (!XAllocColor (display, colormap, &color[i])) {
            while (--i > 0)
                XFreeColors (display, colormap, &color[i].pixel, 1, 0);
            return 0;
            }
        }
    return 1;
}


/*  ======================= TEST CODE ================================= */

    /*  Open a window in the given window and colormap. */
static Window OpenWindow (display, visual, colormap)
    Display                *display;
    Visual                 *visual;
    Colormap                colormap;
{
XSetWindowAttributes        atts;
XVisualInfo                 vinfo, *pVInfo;
int                         depth, nItems;
Window                      window;

    vinfo.visualid = visual->visualid;
    if (!(pVInfo = XGetVisualInfo (display, VisualIDMask, &vinfo, &nItems)))
        return (Window)0;
    depth = pVInfo->depth;

    atts.colormap = colormap;
    atts.background_pixel = atts.border_pixel = 0;      /* black? */
    window = XCreateWindow (display, DefaultRootWindow (display), 0, 0,
                 1, 1, 0, depth, InputOutput, visual, 
                 CWColormap | CWBorderPixel | CWBackPixel, &atts);
    return window;
}


main (argc, argv)
    int		argc;
    char	**argv;
{
char *displayName = "", *name;
Display        *display;
int             i, screen;
int             nWindows;
XGCValues       values;
XEvent         event;

#define         MAX_WINDOWS 20
struct {
    char       *name;
    GC          gc;
    Window      window;
    XscmInfoRec info;
    } windows [MAX_WINDOWS];

#define EVENTS (ButtonPressMask | KeyPressMask)
#define STREQ(str, cmpstr) (strcmp ((str), (cmpstr)) == 0)

    /*  In testimage.c: a test image, size IWIDTH by IHEIGHT. */
#define IWIDTH 250
#define IHEIGHT 300
extern unsigned char testImage[];

    argc--;
    argv++;

    while (argc-- > 0) {
        name = *argv++;
        if (STREQ (name, "-display")) {
            if (argc-- <= 0) {
                fprintf (stderr, "-display requires display name argument.\n");
                exit (1);
	        }
            displayName = *argv++;
	    }
        else {
            fprintf (stderr, "? '%s' is not a valid option\n", name);
            exit (1);
	    }
        }

    display = XOpenDisplay (displayName);
    if (!display) {
        fprintf (stderr, "? Can't open display '%s'.\n", displayName);
        exit (1);
        }

    nWindows = 0;

        /*  Look for explicit Standard ColorMaps (SCM) */
    if (XscmFind24BitSCM (display, 0, &windows[nWindows].info)) {
        windows[nWindows].name = "24 Bit";
        nWindows++;
        }
    if (XscmFind8BitSCM (display, 0, &windows[nWindows].info)) {
        windows[nWindows].name = "8 Bit";
        nWindows++;
        }
    if (XscmFindDefaultSCM (display, 0, &windows[nWindows].info)) {
        windows[nWindows].name = "Default";
        nWindows++;
        }
    if (XscmFindGraySCM (display, 0, &windows[nWindows].info)) {
        windows[nWindows].name = "Gray";
        nWindows++;
        }
    if (XscmFindCRSCM (display, 0, 0, &windows[nWindows].info)) {
        windows[nWindows].name = "TrueColor Color Recovery";
        nWindows++;
        }
    if (XscmFindCRSCM (display, 0, 1, &windows[nWindows].info)) {
        windows[nWindows].name = "PseudoColor Color Recovery";
        nWindows++;
        }

        /*  Look for "best", some of which may end up the same as the
            explicit ones.  Look for "any" (may be non-default visual/colormap);
            "default visual" (but possibly not default colormap) and 
            "default colormap" (both default visual and colormap).
        */
    if (XscmFindColorSCM (display, 0, XSCM_ANY, &windows[nWindows].info)) {
        windows[nWindows].name = "Any";
        nWindows++;
        }

    if (XscmFindColorSCM (display, 0, XSCM_DEFAULT_VISUAL, 
                          &windows[nWindows].info)) {
        windows[nWindows].name = "Default Visual";
        nWindows++;
        }

    if (XscmFindColorSCM (display, 0, XSCM_DEFAULT_COLORMAP, 
                          &windows[nWindows].info)) {
        windows[nWindows].name = "Default Colormap";
        nWindows++;
        }

    if (nWindows <= 0) {
        fprintf (stderr, "No standard colormap windows found.\n");
        exit (1);
        }

        /*  Setup each window found above. */
    for (i = 0; i < nWindows; i++) {
        windows[i].window = OpenWindow (display, windows[i].info.visual, 
                                        windows[i].info.colormap);
        if (!windows[i].window) {
            fprintf (stderr, "? Error creating window for '%s'.\n",
                     windows[i].name);
            exit (1);
            }
        XSelectInput (display, windows[i].window, EVENTS);
        XResizeWindow (display, windows[i].window, IWIDTH, IHEIGHT);
        XMapWindow (display, windows[i].window);
        XStoreName (display, windows[i].window, windows[i].name);
        windows[i].gc = XCreateGC (display, windows[i].window, 0, &values);
        if (!windows[i].gc) {
            fprintf (stderr, "? Error creating gc for '%s'.\n", windows[i].name);
            exit (1);
            }
        }

        /*  Get event; for each event, find window for it and display image. */
    while (1) {
        XNextEvent (display, &event);

        for (i = 0; i < nWindows; i++) {
            if (event.xany.window == windows[i].window) {
                XClearWindow (display, windows[i].window);
                XFlush (display);

                if (!XscmDisplay (&windows[i].info, testImage, 
                      (IWIDTH * 3), IWIDTH, IHEIGHT, 
                      (int *)0, (int *)0, (int *)0, 
                      windows[i].window, windows[i].gc, 0, 0))
                {
                    fprintf (stderr, "? Error displaying for '%s'.\n", 
                                      windows[i].name);
                    exit (1);
	        }
                XFlush (display);
                break;
	        }
	    }
        }

    exit (0);
}

