/*
(c) Copyright 1994 HEWLETT-PACKARD COMPANY

To anyone who acknowledges that this file is provided 
"AS IS" without any express or implied warranty:
permission to use, copy, modify, and distribute this 
file for any purpose is hereby granted without fee, 
provided that the above copyright notice and this 
notice appears in all copies, and that the name of 
Hewlett-Packard Company not be used in advertising or 
publicity pertaining to distribution of the software 
without specific, written prior permission.  Hewlett-
Packard Company makes no representations about the 
suitability of this software for any purpose.
*/



/*
    Color usage program: print out information on X color usage.
    Currently works only for 8 plane DefaultVisual displays.
*/

char versionName[] = "@(#)xcoloruse version 1.0.0";

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

main (argc, argv)
    int		argc;
    char	**argv;
{
Display                *display;
char                   *displayName = "", *name;
Visual                 *visual;
XVisualInfo             template, *pInfo;
int                     screen, notUsed, nAlloc, totalColors, i;
int                     nReadOnly, nWriteable;
unsigned long           pixel, planes;
Colormap                colormap;
#define                 MAX_COLORS 256
XColor                  color;
int                     allocated[MAX_COLORS], readonly[MAX_COLORS];
int                     full = 0;
#define STREQ(str, cmpstr) (strcmp ((str), (cmpstr)) == 0)

    argc--;
    argv++;

    while (argc-- > 0) {
        name = *argv++;
        if (STREQ (name, "-display")) {
            if (argc-- <= 0) {
                fprintf (stderr, "-display requires display name argument.\n");
                exit (1);
	        }
            displayName = *argv++;
	    }
        else if (STREQ (name, "-full"))
            full = 1;
        else {
            fprintf (stderr, "? '%s' is not a valid option\n", name);
            fprintf (stderr, "Usage: xcoloruse [-full] [-display <display>]\n");
            exit (1);
	    }
        }

    display = XOpenDisplay (displayName);
    if (!display) {
        fprintf (stderr, "Can't open Display.\n");
        exit (1);
        }
    screen = DefaultScreen (display);
    visual = DefaultVisual (display, screen);
    colormap = DefaultColormap (display, screen);

    template.visualid = XVisualIDFromVisual (visual);
    pInfo = XGetVisualInfo (display, VisualIDMask, &template, &notUsed);
    if (pInfo->class != PseudoColor) {
        fprintf (stderr, "Default visual not PseudoColor.\n");
        exit (1);
        }

        /*  Allocate writeable colors until error; count same */
    totalColors = pInfo->colormap_size;
    if (totalColors > MAX_COLORS) {
        printf ("? Colormap size = %d colors; can only handle %d.\n",
            totalColors, MAX_COLORS);
        exit (1);
        }

    for (i = 0; i < MAX_COLORS; i++)
        allocated[i] = readonly[i] = 0;
    for (nAlloc = 0; nAlloc < totalColors; nAlloc++) {
        if (!XAllocColorCells (display, colormap, 0, &planes, 0, &pixel, 1))
            break;
        if (pixel > (MAX_COLORS - 1)) {
            printf ("? pixel %d > max - 1; can't handle.\n", pixel);
            exit (1);
	    }
        allocated [pixel] = 1;
        }

        /*  "allocated" now marks the colors allocated writeable which 
            therefore can't be in use; the others are.  Query them and try 
            allocating the same color as read-only.  If the same pixel comes 
            back, assume that color was allocated read-only, else writeable.
        */
    nReadOnly = nWriteable = 0;
    for (i = 0; i < totalColors; i++) {
        if (!allocated[i]) {
            color.pixel = i;
            XQueryColor (display, colormap, &color);
            if (XAllocColor (display, colormap, &color)
             && (color.pixel == i)) {
                 readonly[i] = 1;
                 nReadOnly++;
	         }
            else nWriteable++;
	    }
        }
    

    printf ("    xcoloruse: %d total colors; %d used %d available.\n",
             totalColors, totalColors - nAlloc, nAlloc);
    printf ("               %d of the used are writeable; %d are read-only.\n",
             nWriteable, nReadOnly);

        /*  If full listing, list all used colors and mark the writeable ones */
    if (full) {
        for (i = 0; i < totalColors; i++) {
            printf (" %3d (0x%2.2x): ", i, i);
            if (allocated[i])
                printf ("unused\n");
    	else {
                if (readonly[i])
                     printf ("read-only ");
                else printf ("writeable ");
                color.pixel = i;
                XQueryColor (display, colormap, &color);
                printf ("%5d %5d %5d ", color.red, color.green, color.blue);
                printf ("#%4.4x%4.4x%4.4x\n",color.red,color.green,color.blue);
    	    }
        }
    }
}

